@@ -3,30 +3,54 @@ import urllib
 import json
 
 from .sdk_token_provider import SdkTokenProvider
-from .api_types import FireblocksApiException, TRANSACTION_TYPES, TRANSACTION_STATUS_TYPES, PEER_TYPES, TransferPeerPath, DestinationTransferPeerPath, TransferTicketTerm, TRANSACTION_TRANSFER, SIGNING_ALGORITHM, UnsignedMessage, FEE_LEVEL
+from .api_types import FireblocksApiException, TRANSACTION_TYPES, TRANSACTION_STATUS_TYPES, PEER_TYPES, \
+    TransferPeerPath, DestinationTransferPeerPath, TransferTicketTerm, TRANSACTION_TRANSFER, SIGNING_ALGORITHM, \
+    UnsignedMessage, FEE_LEVEL
 from fireblocks_sdk.api_types import TransactionDestination
 
+
+def handle_response(response, page_mode=False):
+    try:
+        response_data = response.json()
+    except:
+        response_data = None
+    if response.status_code >= 300:
+        if type(response_data) is dict:
+            error_code = response_data.get("code")
+            raise FireblocksApiException("Got an error from fireblocks server: " + response.text, error_code)
+        else:
+            raise FireblocksApiException("Got an error from fireblocks server: " + response.text)
+    else:
+        if page_mode:
+            return {'transactions': response_data,
+                    'pageDetails': {'prevPage': response.headers.get('prev-page', ''),
+                                    'nextPage': response.headers.get('next-page', '')}}
+        return response_data
+
+
 class FireblocksSDK(object):
 
-    def __init__(self, private_key, api_key, api_base_url="https://api.fireblocks.io"):
+    def __init__(self, private_key, api_key, api_base_url="https://api.fireblocks.io", timeout=None):
         """Creates a new Fireblocks API Client.
 
         Args:
             private_key (str): A string representation of your private key (in PEM format)
             api_key (str): Your api key. This is a uuid you received from Fireblocks
             base_url (str): The fireblocks server URL. Leave empty to use the default server
+            timeout (number): Timeout for http requests in seconds
         """
         self.private_key = private_key
         self.api_key = api_key
         self.base_url = api_base_url
         self.token_provider = SdkTokenProvider(private_key, api_key)
+        self.timeout = timeout
 
     def get_supported_assets(self):
         """Gets all assets that are currently supported by Fireblocks"""
 
         return self._get_request("/v1/supported_assets")
 
-    def get_vault_accounts(self, name_prefix=None, name_suffix=None, min_amount_threshold=None):
+    def get_vault_accounts(self, name_prefix=None, name_suffix=None, min_amount_threshold=None, assetId=None):
         """Gets all vault accounts for your tenant
 
         Args:
@@ -47,6 +71,9 @@ class FireblocksSDK(object):
         if min_amount_threshold is not None:
             params['minAmountThreshold'] = min_amount_threshold
 
+        if assetId is not None:
+            params['assetId'] = assetId
+
         if params:
             url = url + "?" + urllib.parse.urlencode(params)
 
@@ -77,6 +104,15 @@ class FireblocksSDK(object):
 
         return self._get_request(f"/v1/vault/accounts/{vault_account_id}/{asset_id}")
 
+    def refresh_vault_asset_balance(self, vault_account_id, asset_id, idempotency_key=None):
+        """Gets a single vault account asset after forcing refresh from the blockchain
+        Args:
+            vault_account_id (string): The id of the requested account
+            asset_id (string): The symbol of the requested asset (e.g BTC, ETH)
+        """
+
+        return self._post_request(f"/v1/vault/accounts/{vault_account_id}/{asset_id}/balance", {}, idempotency_key)
+
     def get_deposit_addresses(self, vault_account_id, asset_id):
         """Gets deposit addresses for an asset in a vault account
         Args:
@@ -95,7 +131,8 @@ class FireblocksSDK(object):
 
         return self._get_request(f"/v1/vault/accounts/{vault_account_id}/{asset_id}/unspent_inputs")
 
-    def generate_new_address(self, vault_account_id, asset_id, description=None, customer_ref_id=None, idempotency_key=None):
+    def generate_new_address(self, vault_account_id, asset_id, description=None, customer_ref_id=None,
+                             idempotency_key=None):
         """Generates a new address for an asset in a vault account
 
         Args:
@@ -106,7 +143,9 @@ class FireblocksSDK(object):
             idempotency_key (str, optional)
         """
 
-        return self._post_request(f"/v1/vault/accounts/{vault_account_id}/{asset_id}/addresses", { "description": description or '', "customerRefId": customer_ref_id or ''}, idempotency_key)
+        return self._post_request(f"/v1/vault/accounts/{vault_account_id}/{asset_id}/addresses",
+                                  {"description": description or '', "customerRefId": customer_ref_id or ''},
+                                  idempotency_key)
 
     def set_address_description(self, vault_account_id, asset_id, address, tag=None, description=None):
         """Sets the description of an existing address
@@ -119,9 +158,11 @@ class FireblocksSDK(object):
             description (string, optional): The description to set, or none for no description
         """
         if tag:
-            return self._put_request(f"/v1/vault/accounts/{vault_account_id}/{asset_id}/addresses/{address}:{tag}", { "description": description or ''})
+            return self._put_request(f"/v1/vault/accounts/{vault_account_id}/{asset_id}/addresses/{address}:{tag}",
+                                     {"description": description or ''})
         else:
-            return self._put_request(f"/v1/vault/accounts/{vault_account_id}/{asset_id}/addresses/{address}", { "description": description or ''})
+            return self._put_request(f"/v1/vault/accounts/{vault_account_id}/{asset_id}/addresses/{address}",
+                                     {"description": description or ''})
 
     def get_network_connections(self):
         """Gets all network connections for your tenant"""
@@ -174,7 +215,8 @@ class FireblocksSDK(object):
             "amount": amount
         }
 
-        return self._post_request(f"/v1/exchange_accounts/{exchange_account_id}/{asset_id}/transfer_to_subaccount", body, idempotency_key)
+        return self._post_request(f"/v1/exchange_accounts/{exchange_account_id}/{asset_id}/transfer_to_subaccount",
+                                  body, idempotency_key)
 
     def transfer_from_subaccount(self, exchange_account_id, subaccount_id, asset_id, amount, idempotency_key=None):
         """Transfer from a subaccount to a main exchange account
@@ -191,7 +233,8 @@ class FireblocksSDK(object):
             "amount": amount
         }
 
-        return self._post_request(f"/v1/exchange_accounts/{exchange_account_id}/{asset_id}/transfer_from_subaccount", body, idempotency_key)
+        return self._post_request(f"/v1/exchange_accounts/{exchange_account_id}/{asset_id}/transfer_from_subaccount",
+                                  body, idempotency_key)
 
     def get_fiat_accounts(self):
         """Gets all fiat accounts for your tenant"""
@@ -236,7 +279,8 @@ class FireblocksSDK(object):
         return self._post_request(f"/v1/fiat_accounts/{account_id}/deposit_from_linked_dda", body, idempotency_key)
 
     def get_transactions_with_page_info(self, before=0, after=None, status=None, limit=None, txhash=None,
-                         assets=None, source_type=None, source_id=None, dest_type=None, dest_id=None, next_or_previous_path=None):
+                                        assets=None, source_type=None, source_id=None, dest_type=None, dest_id=None,
+                                        next_or_previous_path=None):
         """Gets a list of transactions matching the given filters or path.
         Note that "next_or_previous_path" is mutually exclusive with other parameters.
         If you wish to iterate over the nextPage/prevPage pages, please provide only the "next_or_previous_path" parameter from `pageDetails` response
@@ -271,7 +315,8 @@ class FireblocksSDK(object):
             suffix_path = next_or_previous_path[index:length]
             return self._get_request(suffix_path, True)
         else:
-            return self._get_transactions(before, after, status, limit, None, txhash, assets, source_type, source_id, dest_type, dest_id, True)
+            return self._get_transactions(before, after, status, limit, None, txhash, assets, source_type, source_id,
+                                          dest_type, dest_id, True)
 
     def get_transactions(self, before=0, after=0, status=None, limit=None, order_by=None, txhash=None,
                          assets=None, source_type=None, source_id=None, dest_type=None, dest_id=None):
@@ -297,9 +342,11 @@ class FireblocksSDK(object):
                 NETWORK_CONNECTION, COMPOUND
             dest_id (str, optional): Only gets transactions with given dest_id
         """
-        return self._get_transactions(before, after, status, limit, order_by, txhash, assets, source_type, source_id, dest_type, dest_id)
+        return self._get_transactions(before, after, status, limit, order_by, txhash, assets, source_type, source_id,
+                                      dest_type, dest_id)
 
-    def _get_transactions(self, before, after, status, limit, order_by, txhash, assets, source_type, source_id, dest_type, dest_id, page_mode=False):
+    def _get_transactions(self, before, after, status, limit, order_by, txhash, assets, source_type, source_id,
+                          dest_type, dest_id, page_mode=False):
         path = "/v1/transactions"
         params = {}
 
@@ -316,7 +363,7 @@ class FireblocksSDK(object):
             params['limit'] = limit
         if order_by:
             params['orderBy'] = order_by
-        if  txhash:
+        if txhash:
             params['txHash'] = txhash
         if assets:
             params['assets'] = assets
@@ -354,7 +401,6 @@ class FireblocksSDK(object):
         """
         return self._get_request(f"/v1/internal_wallets/{wallet_id}/{asset_id}")
 
-
     def get_external_wallets(self):
         """Gets all external wallets for your tenant"""
 
@@ -394,7 +440,6 @@ class FireblocksSDK(object):
 
         return self._get_request(f"/v1/transactions/external_tx_id/{external_tx_id}")
 
-
     def get_fee_for_asset(self, asset_id):
         """Gets the estimated fees for an asset
 
@@ -404,7 +449,8 @@ class FireblocksSDK(object):
 
         return self._get_request(f"/v1/estimate_network_fee?assetId={asset_id}")
 
-    def estimate_fee_for_transaction(self, asset_id, amount, source, destination=None , tx_type=TRANSACTION_TRANSFER, idempotency_key=None):
+    def estimate_fee_for_transaction(self, asset_id, amount, source, destination=None, tx_type=TRANSACTION_TRANSFER,
+                                     idempotency_key=None, destinations=None):
         """Estimates transaction fee
 
         Args:
@@ -414,13 +460,15 @@ class FireblocksSDK(object):
             amount (str): The amount
             tx_type (str, optional): Transaction type: either TRANSFER, MINT, BURN, TRANSACTION_SUPPLY_TO_COMPOUND or TRANSACTION_REDEEM_FROM_COMPOUND. Default is TRANSFER.
             idempotency_key (str, optional)
+            destinations (list of TransactionDestination objects, optional): For UTXO based assets, send to multiple destinations which should be specified using this field.
         """
 
         if tx_type not in TRANSACTION_TYPES:
             raise FireblocksApiException("Got invalid transaction type: " + tx_type)
 
         if not isinstance(source, TransferPeerPath):
-            raise FireblocksApiException("Expected transaction source of type TransferPeerPath, but got type: " + type(source))
+            raise FireblocksApiException(
+                "Expected transaction source of type TransferPeerPath, but got type: " + type(source))
 
         body = {
             "assetId": asset_id,
@@ -431,9 +479,16 @@ class FireblocksSDK(object):
 
         if destination:
             if not isinstance(destination, (TransferPeerPath, DestinationTransferPeerPath)):
-                raise FireblocksApiException("Expected transaction fee estimation destination of type DestinationTransferPeerPath or TransferPeerPath, but got type: " + type(destination))
+                raise FireblocksApiException(
+                    "Expected transaction fee estimation destination of type DestinationTransferPeerPath or TransferPeerPath, but got type: " + type(
+                        destination))
             body["destination"] = destination.__dict__
 
+        if destinations:
+            if any([not isinstance(x, TransactionDestination) for x in destinations]):
+                raise FireblocksApiException("Expected destinations of type TransactionDestination")
+            body['destinations'] = [dest.__dict__ for dest in destinations]
+
         return self._post_request("/v1/transactions/estimate_fee", body, idempotency_key)
 
     def cancel_transaction_by_id(self, txid, idempotency_key=None):
@@ -444,7 +499,7 @@ class FireblocksSDK(object):
             idempotency_key (str, optional)
         """
 
-        return self._post_request(f"/v1/transactions/{txid}/cancel", idempotency_key)
+        return self._post_request(f"/v1/transactions/{txid}/cancel", idempotency_key=idempotency_key)
 
     def drop_transaction(self, txid, fee_level=None, requested_fee=None, idempotency_key=None):
         """Drops the selected transaction from the blockchain by replacing it with a 0 ETH transaction to itself
@@ -492,7 +547,7 @@ class FireblocksSDK(object):
             vault_account_id (str): The vault account Id
             idempotency_key (str, optional)
         """
-        return self._post_request(f"/v1/vault/accounts/{vault_account_id}/hide", idempotency_key)
+        return self._post_request(f"/v1/vault/accounts/{vault_account_id}/hide", idempotency_key=idempotency_key)
 
     def unhide_vault_account(self, vault_account_id, idempotency_key=None):
         """Returns the vault account to being visible in the web console
@@ -501,7 +556,7 @@ class FireblocksSDK(object):
             vault_account_id (str): The vault account Id
             idempotency_key (str, optional)
         """
-        return self._post_request(f"/v1/vault/accounts/{vault_account_id}/unhide", idempotency_key)
+        return self._post_request(f"/v1/vault/accounts/{vault_account_id}/unhide", idempotency_key=idempotency_key)
 
     def freeze_transaction_by_id(self, txId, idempotency_key=None):
         """Freezes the selected transaction
@@ -510,7 +565,7 @@ class FireblocksSDK(object):
             txId (str): The transaction ID to freeze
             idempotency_key (str, optional)
         """
-        return self._post_request(f"/v1/transactions/{txId}/freeze", idempotency_key)
+        return self._post_request(f"/v1/transactions/{txId}/freeze", idempotency_key=idempotency_key)
 
     def unfreeze_transaction_by_id(self, txId, idempotency_key=None):
         """Unfreezes the selected transaction
@@ -519,7 +574,7 @@ class FireblocksSDK(object):
             txId (str): The transaction ID to unfreeze
             idempotency_key (str, optional)
         """
-        return self._post_request(f"/v1/transactions/{txId}/unfreeze", idempotency_key)
+        return self._post_request(f"/v1/transactions/{txId}/unfreeze", idempotency_key=idempotency_key)
 
     def update_vault_account(self, vault_account_id, name):
         """Updates a vault account.
@@ -543,7 +598,19 @@ class FireblocksSDK(object):
             idempotency_key (str, optional)
         """
 
-        return self._post_request(f"/v1/vault/accounts/{vault_account_id}/{asset_id}", idempotency_key)
+        return self._post_request(f"/v1/vault/accounts/{vault_account_id}/{asset_id}", idempotency_key=idempotency_key)
+
+    def activate_vault_asset(self, vault_account_id, asset_id, idempotency_key=None):
+        """Retry to create a vault asset for a vault asset that failed
+
+        Args:
+            vault_account_id (str): The vault account Id
+            asset_id (str): The symbol of the asset to add (e.g BTC, ETH)
+            idempotency_key (str, optional)
+        """
+
+        return self._post_request(f"/v1/vault/accounts/{vault_account_id}/{asset_id}/activate",
+                                  idempotency_key=idempotency_key)
 
     def set_vault_account_customer_ref_id(self, vault_account_id, customer_ref_id, idempotency_key=None):
         """Sets an AML/KYT customer reference ID for the vault account
@@ -554,9 +621,11 @@ class FireblocksSDK(object):
             idempotency_key (str, optional)
         """
 
-        return self._post_request(f"/v1/vault/accounts/{vault_account_id}/set_customer_ref_id", {"customerRefId": customer_ref_id or ''}, idempotency_key)
+        return self._post_request(f"/v1/vault/accounts/{vault_account_id}/set_customer_ref_id",
+                                  {"customerRefId": customer_ref_id or ''}, idempotency_key)
 
-    def set_vault_account_customer_ref_id_for_address(self, vault_account_id, asset_id, address, customer_ref_id=None, idempotency_key=None):
+    def set_vault_account_customer_ref_id_for_address(self, vault_account_id, asset_id, address, customer_ref_id=None,
+                                                      idempotency_key=None):
         """Sets an AML/KYT customer reference ID for the given address
 
         Args:
@@ -567,10 +636,11 @@ class FireblocksSDK(object):
             idempotency_key (str, optional)
         """
 
-        return self._post_request(f"/v1/vault/accounts/{vault_account_id}/{asset_id}/addresses/{address}/set_customer_ref_id", {"customerRefId": customer_ref_id or ''}, idempotency_key)
+        return self._post_request(
+            f"/v1/vault/accounts/{vault_account_id}/{asset_id}/addresses/{address}/set_customer_ref_id",
+            {"customerRefId": customer_ref_id or ''}, idempotency_key)
 
-
-    def create_external_wallet(self, name, customer_ref_id, idempotency_key=None):
+    def create_external_wallet(self, name, customer_ref_id=None, idempotency_key=None):
         """Creates a new external wallet
 
         Args:
@@ -579,7 +649,8 @@ class FireblocksSDK(object):
             idempotency_key (str, optional)
         """
 
-        return self._post_request("/v1/external_wallets", {"name": name, "customerRefId": customer_ref_id or ''}, idempotency_key)
+        return self._post_request("/v1/external_wallets", {"name": name, "customerRefId": customer_ref_id or ''},
+                                  idempotency_key)
 
     def create_internal_wallet(self, name, customer_ref_id=None, idempotency_key=None):
         """Creates a new internal wallet
@@ -590,7 +661,8 @@ class FireblocksSDK(object):
             idempotency_key (str, optional)
         """
 
-        return self._post_request("/v1/internal_wallets", {"name": name, "customerRefId": customer_ref_id or ''}, idempotency_key)
+        return self._post_request("/v1/internal_wallets", {"name": name, "customerRefId": customer_ref_id or ''},
+                                  idempotency_key)
 
     def create_external_wallet_asset(self, wallet_id, asset_id, address, tag=None, idempotency_key=None):
         """Creates a new asset within an exiting external wallet
@@ -609,7 +681,7 @@ class FireblocksSDK(object):
 
         return self._post_request(
             f"/v1/external_wallets/{wallet_id}/{asset_id}", body, idempotency_key
-            )
+        )
 
     def create_internal_wallet_asset(self, wallet_id, asset_id, address, tag=None, idempotency_key=None):
         """Creates a new asset within an exiting internal wallet
@@ -628,10 +700,13 @@ class FireblocksSDK(object):
 
         return self._post_request(
             f"/v1/internal_wallets/{wallet_id}/{asset_id}", body, idempotency_key
-            )
+        )
 
-
-    def create_transaction(self, asset_id, amount=None, source=None, destination=None, fee=None, gas_price=None, wait_for_status=False, tx_type=TRANSACTION_TRANSFER, note=None, network_fee=None, customer_ref_id=None, replace_tx_by_hash=None, extra_parameters=None, destinations=None, fee_level=None, fail_on_fee=None, max_fee=None, gas_limit=None, idempotency_key=None, external_tx_id=None, treat_as_gross_amount=None):
+    def create_transaction(self, asset_id, amount=None, source=None, destination=None, fee=None, gas_price=None,
+                           wait_for_status=False, tx_type=TRANSACTION_TRANSFER, note=None, network_fee=None,
+                           customer_ref_id=None, replace_tx_by_hash=None, extra_parameters=None, destinations=None,
+                           fee_level=None, fail_on_low_fee=None, max_fee=None, gas_limit=None, idempotency_key=None,
+                           external_tx_id=None, treat_as_gross_amount=None, force_sweep=None):
         """Creates a new transaction
 
         Args:
@@ -644,19 +719,18 @@ class FireblocksSDK(object):
             wait_for_status (bool, optional): If true, waits for transaction status. Default is false.
             tx_type (str, optional): Transaction type: either TRANSFER, MINT, BURN, TRANSACTION_SUPPLY_TO_COMPOUND or TRANSACTION_REDEEM_FROM_COMPOUND. Default is TRANSFER.
             note (str, optional): A custome note that can be associated with the transaction.
-            cpu_staking (number, optional): Cpu stake for EOS transfer.
-            network_staking (number, optional): Network stake for EOS transfer.
-            auto_staking: (boolean, optional): Auto stake for EOS transfer. Staking will be managed by fireblocks.
+            network_fee (str, optional): Transaction blockchain fee (For Ethereum, you can't pass gasPrice, gasLimit and networkFee all together)
             customer_ref_id (string, optional): The ID for AML providers to associate the owner of funds with transactions
             extra_parameters (object, optional)
             destinations (list of TransactionDestination objects, optional): For UTXO based assets, send to multiple destinations which should be specified using this field.
             fee_level (FeeLevel, optional): Transaction fee level: either HIGH, MEDIUM, LOW.
-            fail_on_fee (bool, optional): False by default, if set to true and MEDIUM fee level is higher than the one specified in the transaction, the transction will fail.
+            fail_on_low_fee (bool, optional): False by default, if set to true and MEDIUM fee level is higher than the one specified in the transaction, the transction will fail.
             max_fee (str, optional): The maximum fee (gas price or fee per byte) that should be payed for the transaction.
             gas_limit (number, optional): For ETH-based assets only.
             idempotency_key (str, optional)
             external_tx_id (str, optional): A unique key for transaction provided externally
             treat_as_gross_amount (bool, optional): Determine if amount should be treated as gross or net
+            force_sweep (bool, optional): Determine if transaction should be treated as a forced sweep
         """
 
         if tx_type not in TRANSACTION_TYPES:
@@ -664,7 +738,8 @@ class FireblocksSDK(object):
 
         if source:
             if not isinstance(source, TransferPeerPath):
-                raise FireblocksApiException("Expected transaction source of type TransferPeerPath, but got type: " + type(source))
+                raise FireblocksApiException(
+                    "Expected transaction source of type TransferPeerPath, but got type: " + type(source))
 
         body = {
             "waitForStatus": wait_for_status,
@@ -691,8 +766,8 @@ class FireblocksSDK(object):
         if max_fee:
             body["maxFee"] = max_fee
 
-        if fail_on_fee:
-            body["failOnFee"] = fail_on_fee
+        if fail_on_low_fee:
+            body["failOnLowFee"] = fail_on_low_fee
 
         if gas_price:
             body["gasPrice"] = str(gas_price)
@@ -705,7 +780,9 @@ class FireblocksSDK(object):
 
         if destination:
             if not isinstance(destination, (TransferPeerPath, DestinationTransferPeerPath)):
-                raise FireblocksApiException("Expected transaction destination of type DestinationTransferPeerPath or TransferPeerPath, but got type: " + type(destination))
+                raise FireblocksApiException(
+                    "Expected transaction destination of type DestinationTransferPeerPath or TransferPeerPath, but got type: " + type(
+                        destination))
             body["destination"] = destination.__dict__
 
         if network_fee:
@@ -732,6 +809,9 @@ class FireblocksSDK(object):
         if external_tx_id:
             body["externalTxId"] = external_tx_id
 
+        if force_sweep:
+            body["forceSweep"] = force_sweep
+
         return self._post_request("/v1/transactions", body, idempotency_key)
 
     def delete_internal_wallet(self, wallet_id):
@@ -781,7 +861,8 @@ class FireblocksSDK(object):
             idempotency_key (str, optional)
         """
 
-        return self._post_request(f"/v1/internal_wallets/{wallet_id}/set_customer_ref_id", {"customerRefId": customer_ref_id or ''}, idempotency_key)
+        return self._post_request(f"/v1/internal_wallets/{wallet_id}/set_customer_ref_id",
+                                  {"customerRefId": customer_ref_id or ''}, idempotency_key)
 
     def set_customer_ref_id_for_external_wallet(self, wallet_id, customer_ref_id=None, idempotency_key=None):
         """Sets an AML/KYT customer reference ID for the specific external wallet
@@ -792,7 +873,8 @@ class FireblocksSDK(object):
             idempotency_key (str, optional)
         """
 
-        return self._post_request(f"/v1/external_wallets/{wallet_id}/set_customer_ref_id", {"customerRefId": customer_ref_id or ''}, idempotency_key)
+        return self._post_request(f"/v1/external_wallets/{wallet_id}/set_customer_ref_id",
+                                  {"customerRefId": customer_ref_id or ''}, idempotency_key)
 
     def get_transfer_tickets(self):
         """Gets all transfer tickets of your tenant"""
@@ -851,7 +933,7 @@ class FireblocksSDK(object):
             idempotency_key (str, optional)
         """
 
-        return self._post_request(f"/v1/transfer_tickets/{ticket_id}/cancel", idempotency_key)
+        return self._post_request(f"/v1/transfer_tickets/{ticket_id}/cancel", idempotency_key=idempotency_key)
 
     def execute_ticket_term(self, ticket_id, term_id, source=None, idempotency_key=None):
         """Initiate a transfer ticket transaction
@@ -866,10 +948,10 @@ class FireblocksSDK(object):
 
         if source:
             if not isinstance(source, TransferPeerPath):
-                raise FireblocksApiException("Expected ticket term source Of type TransferPeerPath, but got type: " + type(source))
+                raise FireblocksApiException(
+                    "Expected ticket term source Of type TransferPeerPath, but got type: " + type(source))
             body["source"] = source.__dict__
 
-
         return self._post_request(f"/v1/transfer_tickets/{ticket_id}/{term_id}/transfer", body, idempotency_key)
 
     def set_confirmation_threshold_for_txid(self, txid, required_confirmations_number, idempotency_key=None):
@@ -937,23 +1019,68 @@ class FireblocksSDK(object):
 
         return self._get_request(url)
 
-    def get_gas_station_info(self):
-        "Get configuration and status of the Gas Station account"
+    def allocate_funds_to_private_ledger(self, vault_account_id, asset, allocation_id, amount,
+                                         treat_as_gross_amount=None, idempotency_key=None):
+        """Allocate funds from your default balance to a private ledger
+
+        Args:
+            vault_account_id (string)
+            asset (string)
+            allocation_id (string)
+            amount (string)
+            treat_as_gross_amount (bool, optional)
+            idempotency_key (string, optional)
+        """
+
+        url = f"/v1/vault/accounts/{vault_account_id}/{asset}/lock_allocation"
+
+        return self._post_request(url, {"allocationId": allocation_id, "amount": amount,
+                                        "treatAsGrossAmount": treat_as_gross_amount or False}, idempotency_key)
+
+    def deallocate_funds_from_private_ledger(self, vault_account_id, asset, allocation_id, amount,
+                                             idempotency_key=None):
+        """deallocate funds from a private ledger to your default balance
+
+        Args:
+            vault_account_id (string)
+            asset (string)
+            allocation_id (string)
+            amount (string)
+            idempotency_key (string, optional)
+        """
+
+        url = f"/v1/vault/accounts/{vault_account_id}/{asset}/release_allocation"
+
+        return self._post_request(url, {"allocationId": allocation_id, "amount": amount}, idempotency_key)
+
+    def get_gas_station_info(self, asset_id=None):
+        """Get configuration and status of the Gas Station account"
+        
+        Args:
+            asset_id (string, optional)
+        """
 
         url = f"/v1/gas_station"
 
+        if asset_id:
+            url = url + "/{asset_id}"
+
         return self._get_request(url)
 
-    def set_gas_station_configuration(self, gas_threshold, gas_cap, max_gas_price):
+    def set_gas_station_configuration(self, gas_threshold, gas_cap, max_gas_price=None, asset_id=None):
         """Set configuration of the Gas Station account
 
         Args:
             gasThreshold (str)
             gasCap (str)
             maxGasPrice (str, optional)
+            asset_id (str, optional)
         """
 
         url = f"/v1/gas_station/configuration"
+        
+        if asset_id:
+            url = url + "/{asset_id}"
 
         body = {
             "gasThreshold": gas_threshold,
@@ -1008,15 +1135,17 @@ class FireblocksSDK(object):
             note (str, optional): A custome note that can be associated with the transaction
         """
 
-        if raw_message.algorithm not in SIGNING_ALGORITHM:
-            raise Exception("Got invalid signing algorithm type: " + raw_message.algorithm)
+        if asset_id is None:
+            if raw_message.algorithm not in SIGNING_ALGORITHM:
+                raise Exception("Got invalid signing algorithm type: " + raw_message.algorithm)
 
         if not all([isinstance(x, UnsignedMessage) for x in raw_message.messages]):
             raise FireblocksApiException("Expected messages of type UnsignedMessage")
 
         raw_message.messages = [message.__dict__ for message in raw_message.messages]
 
-        return self.create_transaction(asset_id, source=source, tx_type="RAW", extra_parameters={"rawMessageData": raw_message.__dict__}, note=note)
+        return self.create_transaction(asset_id, source=source, tx_type="RAW",
+                                       extra_parameters={"rawMessageData": raw_message.__dict__}, note=note)
 
     def get_max_spendable_amount(self, vault_account_id, asset_id, manual_signing=False):
         """Get max spendable amount per asset and vault.
@@ -1060,6 +1189,21 @@ class FireblocksSDK(object):
 
         return self._post_request("/v1/webhooks/resend")
 
+    def resend_transaction_webhooks_by_id(self, tx_id, resend_created, resend_status_updated):
+        """Resend webhooks of transaction
+
+        Args:
+            tx_id (str): The transaction for which the message is sent.
+            resend_created (boolean): If true, a webhook will be sent for the creation of the transaction.
+            resend_status_updated (boolean): If true, a webhook will be sent for the status of the transaction.
+        """
+        body = {
+            "resendCreated": resend_created,
+            "resendStatusUpdated": resend_status_updated
+        }
+
+        return self._post_request(f"/v1/webhooks/resend/{tx_id}", body)
+
     def get_users(self):
         """Gets all users of your tenant"""
 
@@ -1067,6 +1211,35 @@ class FireblocksSDK(object):
 
         return self._get_request(url)
 
+    def get_off_exchanges(self):
+        """
+        Get your connected off exchanges virtual accounts
+        """
+        url = f"/v1/off_exchange_accounts"
+
+        return self._get_request(url)
+
+    def get_off_exchange_by_id(self, off_exchange_id):
+        """
+        Get your connected off exchange by it's ID
+        :param off_exchange_id: ID of the off exchange virtual account
+        :return: off exchange entity
+        """
+
+        url = f"/v1/off_exchange_accounts/{off_exchange_id}"
+
+        return self._get_request(url)
+
+    def settle_off_exchange_by_id(self, off_exchange_id):
+        """
+        Create a settle request to your off exchange by it's ID
+        :param off_exchange_id: ID of the off exchange virtual account
+        """
+
+        url = f"/v1/off_exchanges/{off_exchange_id}/settle"
+
+        return self._post_request(url)
+
     def _get_request(self, path, page_mode=False):
         token = self.token_provider.sign_jwt(path)
         headers = {
@@ -1074,15 +1247,8 @@ class FireblocksSDK(object):
             "Authorization": f"Bearer {token}"
         }
 
-        response = requests.get(self.base_url + path, headers=headers)
-        response_data = response.json()
-        if response.status_code >= 300:
-            raise FireblocksApiException("Got an error from fireblocks server: " + response.text)
-        else:
-            if page_mode:
-                return {'transactions': response_data, 'pageDetails': {'prevPage': response.headers.get('prev-page', ''), 'nextPage': response.headers.get('next-page', '')}}
-            else:
-                return response_data
+        response = requests.get(self.base_url + path, headers=headers, timeout=self.timeout)
+        return handle_response(response, page_mode)
 
     def _delete_request(self, path):
         token = self.token_provider.sign_jwt(path)
@@ -1091,11 +1257,8 @@ class FireblocksSDK(object):
             "Authorization": f"Bearer {token}"
         }
 
-        response = requests.delete(self.base_url + path, headers=headers)
-        if response.status_code >= 300:
-            raise FireblocksApiException("Got an error from fireblocks server: " + response.text)
-        else:
-            return response.json()
+        response = requests.delete(self.base_url + path, headers=headers, timeout=self.timeout)
+        return handle_response(response)
 
     def _post_request(self, path, body={}, idempotency_key=None):
         token = self.token_provider.sign_jwt(path, body)
@@ -1111,11 +1274,8 @@ class FireblocksSDK(object):
                 "Idempotency-Key": idempotency_key
             }
 
-        response = requests.post(self.base_url + path, headers=headers, json=body)
-        if response.status_code >= 300:
-            raise FireblocksApiException("Got an error from fireblocks server: " + response.text)
-        else:
-            return response.json()
+        response = requests.post(self.base_url + path, headers=headers, json=body, timeout=self.timeout)
+        return handle_response(response)
 
     def _put_request(self, path, body={}):
         token = self.token_provider.sign_jwt(path, body)
@@ -1125,14 +1285,5 @@ class FireblocksSDK(object):
             "Content-Type": "application/json"
         }
 
-        response = requests.put(self.base_url + path, headers=headers, data=json.dumps(body))
-        if response.status_code >= 300:
-            raise FireblocksApiException("Got an error from fireblocks server: " + response.text)
-        else:
-            return response.json()
-
-
-
-
-
-
+        response = requests.put(self.base_url + path, headers=headers, data=json.dumps(body), timeout=self.timeout)
+        return handle_response(response)
